from __future__ import annotations
import os
from datetime import datetime
from tweepy.models import Status
from functools import cached_property
from typing import Union, List, Optional, Dict


class InstaPost:

    """Minimalistic API response wrapper for an Instagram post"""

    def __init__(self, data: dict, client: Optional["InstaClient"] = None):
        """Initialize an :class:`~InstaPost`

        :param data: the JSON response data of a single Instagram post, found within the :attr:`~.InstaUser.user_data`
        """
        #: Source data from API response
        self.json = data
        self.client = client
        #: The post id
        self.id = data['id']
        self.dimensions: dict = data.get('dimensions', {})
        self.is_video: bool = data.get('is_video', False)
        self.video_url = data.get('video_url', '')
        #: Path of downloaded media, set by :meth:`~.InstaClient.download_post`
        self.filepath: str = ''
        #: Limited data from a successful tweet based off this post, set by :meth:`~.TweetClient.send_tweet`
        self.tweet_data: dict = {}

    def __str__(self):
        return f'Post {self.id} by @{self.owner["username"]} on {self.timestamp}'

    @cached_property
    def children(self) -> List[InstaPost]:
        """If the post is a carousel, returns a list of child :class:`InstaPost`'s"""
        if self.is_carousel:
            edges = self.json['edge_sidecar_to_children']['edges']
            return [InstaPost(edge['node']) for edge in edges]
        return []

    @property
    def permalink(self) -> str:
        return f'https://www.instagram.com/p/{self.shortcode}'

    @property
    def shortcode(self) -> str:
        return self.json.get('shortcode', self.json.get('code', ''))

    @property
    def caption(self) -> str:
        if caption_edge := self.json.get('edge_media_to_caption', {}).get('edges', []):
            return caption_edge[0].get('node', {}).get('text', '')
        return ''

    @property
    def likes(self) -> Optional[int]:
        return self.json.get('edge_liked_by', {}).get('count')

    @property
    def media_url(self) -> str:
        """The direct URL to the actual post content

        :returns: the :attr:`~.video_url` if the post is a video, otherwise the :attr:`~.thumbnail_url`
        """
        return self.video_url if self.is_video else self.thumbnail_url

    @property
    def thumbnail_url(self) -> str:
        return self.json.get('display_url',
                             self.json.get('thumbnail_src',
                                           self.json.get('thumbnail_resources',
                                                         [{}])[-1].get('src', '')))

    @property
    def is_downloaded(self) -> bool:
        """Checks the :attr:`~filepath` to see if the post has been downloaded yet"""
        if self.is_carousel:
            return all(child.is_downloaded for child in self.children)
        else:
            return os.path.exists(self.filepath)

    @property
    def is_carousel(self) -> bool:
        return 'edge_sidecar_to_children' in self.json

    @property
    def filename(self) -> str:
        """Concatenates :attr:`~id` + :attr:`~filetype` to create the default filename, for use when saving the post

        **For Example**::

         >> print(post.filename)
         "2868062811604347946.mp4"

        """
        return self.id + self.filetype

    @property
    def filetype(self) -> str:
        """Filetype of the post, based on the value of :attr:`~is_video`"""
        return '.mp4' if self.is_video else '.jpg'

    @property
    def owner(self) -> Dict:
        if owner := self.json.get('owner', {}):
            if not owner.get('username'):
                if self.client and (uid := owner.get('id')):
                    owner['username'] = self.client.get_username(uid)
            return owner
        return dict.fromkeys(['id', 'username'])

    @property
    def timestamp(self) -> Union[datetime, str]:
        if timestamp := self.json.get('taken_at_timestamp', self.json.get('taken_at', '')):
            return datetime.utcfromtimestamp(timestamp)
        return ''

    def add_tweet_data(self, tweet: Status) -> bool:
        """Used by :class:`~.TweetClient` to add minimal tweet data after the post has been tweeted

        :param tweet: a :class:`~tweepy.models.Status` object from a successfully sent tweet
        """
        self.tweet_data = {
            'id': tweet.id_str,
            'created_at': str(tweet.created_at),
            'link': f'https://twitter.com/{tweet.user.screen_name}/status/{tweet.id}',
            'entities': tweet.entities,
            'text': tweet.text
        }
        return True



from . import utils, db
from .db import DBConnection
from .instapost import InstaPost
from .instapage import InstaPage, InstaUser, Hashtag
from .instaclient import InstaClient, USER_AGENT
from .tweetclient import TweetClient
from .profile import Profile
from .instatweet import InstaTweet

__version__ = "v2.2.0"




from __future__ import annotations
import os
import copy
import json
import pickle
from pathlib import Path
from typing import Iterable, Dict
from . import TweetClient, DBConnection, USER_AGENT


class Profile:

    """The :class:`Profile` is a configuration class used extensively throughout the package

    It consists of a :attr:`~page_map` and an associated collection of API/web scraping :ref:`settings <settings>`

    ...

    .. admonition:: About the Page Map
        :class: instatweet

        **The** :attr:`~.page_map` **is a dict containing info about the pages added to a** :class:`~.Profile`

        * It's used to help detect new posts and compose tweets on a per-page basis
        * Entries are created when you :meth:`~.add_pages`, which map the page to a :attr:`~.PAGE_MAPPING`
        * The :attr:`~.PAGE_MAPPING` maintains lists of hashtags, scraped posts, and sent tweets
        * The mapping is updated when you :meth:`~.add_hashtags` and successfully :meth:`~.send_tweet`

    ...

    **[Optional]**

    A unique, identifying :attr:`~name` can optionally be assigned to the Profile,
    which may then be used to :meth:`~save` and :meth:`~load` its settings

    The save location is determined by the value of :attr:`Profile.local` as follows:

    * If ``True``, saves are made locally to the :attr:`~LOCAL_DIR` as .pickle files
    * If ``False``, saves are made remotely to a database as pickle bytes

    See :ref:`save-profile` for more information

    ...
    """
    #: Template for an entry in the :attr:`~page_map`
    PAGE_MAPPING: Dict = {'hashtags': [], 'scraped': [], 'tweets': []}
    #: Directory where local profiles are saved
    LOCAL_DIR: str = Path(__file__).parent.parent.joinpath("profiles")

    def __init__(self, name: str = 'default', local: bool = True, **kwargs):
        """Create a new :class:`Profile`

        :param name: unique profile name
        :param local: indicates if profile is being saved locally or on a remote database
        :param kwargs: see below

        :Keyword Arguments:
            * *session_id* (``str``)
                Instagram ``sessionid`` cookie, obtained by logging in through browser
            * *twitter_keys* (``dict``)
                Twitter API Keys with v1.1 endpoint access
                (see :attr:`~.TweetClient.DEFAULT_KEYS` for a template)
            * *user_agent* (``str``) -- Optional
                The user agent to use for requests
            * *proxy_key* (``str``) -- Optional
                Environment variable to retrieve proxies from
        """
        self.local = local
        self.name = name    # Will raise Exception if name is already used

        #: Instagram ``sessionid`` cookie, obtained by logging in through browser
        self.session_id: str = kwargs.get('session_id', '')
        #: Twitter API Keys with v1.1 endpoint access (see :attr:`~.DEFAULT_KEYS` for a template)
        self.twitter_keys: Dict = kwargs.get('twitter_keys', TweetClient.DEFAULT_KEYS)
        #: The user agent to use for requests
        self.user_agent: str = kwargs.get('user_agent', USER_AGENT)
        #: Environment variable to retrieve proxies from
        self.proxy_key: str = kwargs.get('proxy_key', None)
        #: Mapping of added Instagram pages and their :attr:`~PAGE_MAPPING`
        self.page_map: Dict[str, Dict] = kwargs.get('page_map', {})

    @classmethod
    def load(cls, name: str, local: bool = True) -> Profile:
        """Loads an existing profile from a locally saved pickle file or remotely stored pickle bytes

        :param name: the name of the :class:`Profile` to load
        :param local: whether the profile is saved locally (default, ``True``) or remotely on a database
        """
        if not cls.profile_exists(name, local):
            raise LookupError(
                f'No {"local" if local else "database"} profile found with the name "{name}"'
            )
        if local:
            with open(cls.get_local_path(name), 'rb') as f:
                return pickle.load(f)
        else:
            with DBConnection() as db:
                return db.load_profile(name)

    @classmethod
    def from_json(cls, json_str: str) -> Profile:
        """Creates a profile from a JSON formatted string of config settings"""
        return cls.from_dict(json.loads(json_str))

    @classmethod
    def from_dict(cls, d: Dict) -> Profile:
        """Creates a profile from a dictionary of config settings"""
        return cls(**d)

    @staticmethod
    def profile_exists(name: str, local: bool = True) -> bool:
        """Checks locally/remotely to see if a :class:`~Profile` with the specified name has an existing save file

        Whenever the :attr:`~name` is changed, its property setter calls this method to ensure
        you don't accidentally overwrite a save that already :attr:`~exists`

        :param name: the name of the :class:`Profile` to check for
        :param local: the location (local/remote) to check for an existing save

        """
        if local:
            return os.path.exists(Profile.get_local_path(name))
        else:
            with DBConnection() as db:
                return bool(db.query_profile(name).first())

    @staticmethod
    def get_local_path(name: str) -> str:
        """Returns filepath of where a local profile would be saved"""
        return os.path.join(Profile.LOCAL_DIR, name) + '.pickle'

    def add_pages(self, pages: Iterable, send_tweet: bool = False) -> None:
        """Add Instagram page(s) to the :attr:`~.page_map` for subsequent monitoring

        * An Instagram profile can be added as ``"@username"`` or ``"username"``
        * A hashtag must be added as ``"#hashtag"``


        .. note:: By default, newly added pages won't have their posts tweeted the first time they're scraped

           * The IDs of the most recent posts are stored in the ``scraped`` list
           * Any new posts from that point forward will be tweeted

           You can scrape AND tweet posts on the first run by setting ``send_tweet=True``


        :param pages: Instagram pages to automatically scrape and tweet content from
        :param send_tweet: choose if tweets should be sent on the first scrape, or only for new posts going forward
        """
        if not isinstance(pages, Iterable):
            raise TypeError(f'Invalid type provided. `pages` must be an Iterable')
        if isinstance(pages, str):
            pages = [pages]

        for page in pages:
            mapping = copy.deepcopy(Profile.PAGE_MAPPING)
            self.page_map.setdefault(page.lstrip("@"), mapping)

            if send_tweet:  # Non-empty scraped list will trigger Tweets to send
                self.get_scraped_from(page).append(-1)

            print(f'Added Instagram page {page} to the page map')

        if self.exists:
            self._save_profile(alert=False)

    def add_hashtags(self, page: str, hashtags: Iterable):
        """Add hashtag(s) to a page in the :attr:`~.page_map`, which will be randomly chosen from when composing Tweets

        :param page: the page in the page map to add hashtags to
        :param hashtags: hashtags to choose from and include in any Tweets where content comes from this page
        """
        if not isinstance(hashtags, Iterable):
            raise TypeError("Hashtags must be provided as a string or iterable of strings")
        if isinstance(hashtags, str):
            hashtags = [hashtags]

        tags = self.get_hashtags_for(page)  # Retrieve the current hashtag list
        tags.extend(set(hashtags) - set(tags))  # Add new ones (case-sensitive)

        if self.exists:
            self._save_profile(alert=False)
        print(f'Added hashtags for {page}')

    def save(self, name: str = None, alert: bool = True) -> bool:
        """Pickles and saves the :class:`Profile` using the specified or currently set name.

        :param name: name to save the :class:`Profile` under; replaces the current :attr:`~.name`
        :param alert: set to ``True`` to print a message upon successful save
        """
        if name:
            self.name = name
        if self.is_default:  # Profile name wasn't specified and wasn't previously set
            raise AttributeError('Profile name is required to save the profile')
        else:
            return self._save_profile(alert=alert)

    def _save_profile(self, alert: bool = True) -> bool:
        """Internal function to save the profile, based on the value of :attr:`~.local`"""
        if self.local:
            with open(self.profile_path, 'wb') as f:
                pickle.dump(self, f)
            if alert:
                print(f'Saved Local Profile {self.name}')
            return True
        else:
            with DBConnection() as db:
                return db.save_profile(profile=self, alert=alert)

    def validate(self) -> None:
        """Checks to see if the Profile is fully configured for InstaTweeting

        :raises ValueError: if the :attr:`~.session_id`, :attr:`~.twitter_keys`, or :attr:`~.page_map` are invalid
        """
        if not self.session_id:
            raise ValueError('Instagram sessionid cookie is required to scrape posts')

        if bad_keys := [key for key, value in self.twitter_keys.items() if value == 'string']:
            raise ValueError(f'Values not set for the following Twitter keys: {bad_keys}')

        if not self.page_map:
            raise ValueError('You must add at least one Instagram page to auto-tweet from')

    def to_pickle(self) -> bytes:
        """Serializes profile to a pickled byte string"""
        return pickle.dumps(self)

    def to_json(self) -> str:
        """Serializes profile to a JSON formatted string"""
        return json.dumps(self.to_dict())

    def to_dict(self) -> dict:
        """Serializes profile to a dict"""
        return self.config

    def view_config(self):
        """Prints the :attr:`~.config` dict to make it legible"""
        for k, v in self.config.items():
            print(f'{k} : {v}')

    @property
    def config(self) -> dict:
        """Returns a dictionary containing important configuration settings"""
        return {
            'name': self.name,
            'local': self.local,
            'session_id': self.session_id,
            'twitter_keys': self.twitter_keys,
            'user_agent': self.user_agent,
            'proxy_key': self.proxy_key,
            'page_map': self.page_map,
        }

    @property
    def exists(self) -> bool:
        """Returns True if a local save file or database record exists for the currently set profile name"""
        return self.profile_exists(name=self.name, local=self.local)

    @property
    def is_default(self) -> bool:
        """Check if profile :attr:`~.name` is set or not"""
        return self.name == 'default'

    @property
    def profile_path(self) -> str:
        """If :attr:`~.local` is ``True``, returns the file path for where this profile would be/is saved"""
        if self.local and not self.is_default:
            return Profile.get_local_path(self.name)
        return ''

    def get_page(self, page: str) -> dict:
        """Returns the specified page's dict entry in the :attr:`page_map`"""
        return self.page_map[page.lstrip('@')]

    def get_scraped_from(self, page: str) -> list:
        """Returns a list of posts that have been scraped from the specified page"""
        return self.get_page(page)['scraped']

    def get_tweets_for(self, page: str) -> list:
        """Returns a list of tweets that use the specified page's scraped content"""
        return self.get_page(page)['tweets']

    def get_hashtags_for(self, page: str) -> list:
        """Returns the hashtag list for the specified page"""
        return self.get_page(page)['hashtags']

    @property
    def local(self) -> bool:
        """Indicates if saves should be made locally (``True``) or on a remote database (``False``)
        """
        return self._local

    @local.setter
    def local(self, local: bool):
        if local:
            if not os.path.exists(self.LOCAL_DIR):
                os.mkdir(self.LOCAL_DIR)

        self._local = local

    @property
    def name(self) -> str:
        """A name for the Profile

        The :attr:`~Profile.name` is used differently depending on the value of :attr:`~.local`

        * ``local==True``: the name determines the :attr:`~.profile_path` (path where it would save to)
        * ``local==False``: the name is used as the primary key in the :class:`~.Profiles` database table

        ...

        .. admonition:: Profile Names Must Be Unique
            :class: instatweet

            When you set or change the :attr:`~.name`, a property setter will make sure no
            :meth:`~profile_exists` with that name before actually updating it

            * This ensures that you don't accidentally overwrite a different Profile's save data

        ...

        :raises FileExistsError: if :attr:`~local` ``==True`` and a save is found in the :attr:`~LOCAL_DIR`
        :raises ResourceWarning: if :attr:`~local` ``==False`` and a database row is found by :meth:`~.query_profile`
        """
        return self._name

    @name.setter
    def name(self, profile_name: str):
        """Sets the profile name, if a profile with that name doesn't already exist locally/remotely"""
        if profile_name != 'default' and self.profile_exists(profile_name, local=self.local):
            if self.local:
                raise FileExistsError(
                    'Local save file already exists for profile named "{}"\n'.format(profile_name) +
                    'Please choose another name, load the profile, or delete the file.')
            else:
                raise ResourceWarning(
                    'Database record already exists for profile named "{}"\n'.format(profile_name) +
                    'Please choose another name or use InstaTweet.db to load/delete the profile'
                )
        self._name = profile_name

    @property
    def session_id(self) -> str:
        """Instagram ``sessionid`` cookie, obtained by logging in through a browser

            :Tip: If you log into your account with a browser you don't use, the session cookie will last longer
        """
        return self._session_id

    @session_id.setter
    def session_id(self, session_id: str):
        if not isinstance(session_id, str):
            raise TypeError(
                f'Session ID cookie must be of type {str}'
            )
        self._session_id = session_id
        if self.exists:
            self._save_profile(alert=False)

    @property
    def twitter_keys(self) -> Dict:
        """Twitter developer API keys with v1.1 endpoint access. See :attr:`~.DEFAULT_KEYS`"""
        return self._twitter_keys

    @twitter_keys.setter
    def twitter_keys(self, api_keys: dict):
        if not isinstance(api_keys, dict):
            raise TypeError(
                f'Twitter Keys must be of type {dict}'
            )
        if missing_keys := [key for key in TweetClient.DEFAULT_KEYS if key not in api_keys]:
            raise KeyError(
                f'Missing Twitter Keys: {missing_keys}'
            )
        for key in TweetClient.DEFAULT_KEYS:
            if not bool(api_keys[key]):
                raise ValueError(
                    f'Missing Value for Twitter Key: {key}'
                )
        self._twitter_keys = api_keys
        if self.exists:
            self._save_profile(alert=False)



import os
import requests
from requests import Response
from typing import Type, Union, Optional, Dict
from json.decoder import JSONDecodeError
from . import InstaPage, InstaUser, InstaPost, Hashtag


USER_AGENT = "Mozilla/5.0 (Linux; Android 9; GM1903 Build/PKQ1.190110.001; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/75.0.3770.143 Mobile Safari/537.36 Instagram 103.1.0.15.119 Android (28/9; 420dpi; 1080x2260; OnePlus; GM1903; OnePlus7; qcom; sv_SE; 164094539)"
"""Hardcoded user agent proven to work with the :meth:`~.InstaClient.get_user` endpoint
 
 :Version Added: ``v2.0.0b13``
 """


class InstaClient:

    """Minimalistic class for scraping/downloading Instagram user/media data"""

    DOWNLOAD_DIR = os.path.abspath('downloads')  #: [*Optional*] -- Directory to temporarily download media to

    def __init__(self, session_id: str, user_agent: str = USER_AGENT, proxies: Optional[Dict] = None):
        """Initialize an :class:`~InstaClient` with an Instagram sessionid cookie (at minimum)

        .. note:: As of v2.0.0b13, the endpoint used by :meth:`~get_user` seems to require a specific :attr:`~USER_AGENT`
           You can override the hardcoded one if you'd like, but you'll likely get a ``"useragent mismatch"`` response

        :param session_id: valid Instagram sessionid cookie from a browser
        :param user_agent: user agent to use in requests made by the class
        :param proxies: proxies to use in requests made by the class
        """
        if not isinstance(session_id, str):
            raise TypeError('session_id must be a string')

        self.session_id = session_id
        self.user_agent = user_agent
        self.proxies = proxies

        if not os.path.exists(InstaClient.DOWNLOAD_DIR):
            os.mkdir(InstaClient.DOWNLOAD_DIR)

    def request(self, url: str) -> requests.Response:
        """Sends a request using the :attr:`cookies`, :attr:`headers`, and :attr:`proxies`

        :param url: the Instagram URL to send the request to
        """
        return requests.get(
            url, headers=self.headers,
            cookies=self.cookies,
            proxies=self.proxies
        )

    def scrape(self, page: str) -> InstaPage:
        """Scrapes an Instagram page and wraps the response data

        :param page: an Instagram hashtag (prefixed with ``#``) or username
        :returns: an :class:`~.InstaUser` or :class:`~.Hashtag`
        """
        if isinstance(page, str):
            if page.startswith("#"):
                return self.get_hashtag(page)
            return self.get_user(page)
        raise TypeError(f"`page` must be of type {str}")

    def get_hashtag(self, tag: str, max_id: str = '') -> Hashtag:
        """Scrapes an Instagram hashtag and wraps the response with :class:`~.Hashtag`

        :param tag: the hashtag to scrape (with or without a ``#``)
        :param max_id: the end cursor
        """
        tag = tag.lstrip("#")
        endpoint = f'https://www.instagram.com/explore/tags/{tag}/?__a=1&max_id={max_id}&__d=dis'
        response = self.request(endpoint)
        return self._wrap(tag, response, Hashtag)

    def get_user(self, username: str) -> InstaUser:
        """Scrapes an Instagram user's profile and wraps the response with :class:`~.InstaUser`

        :param username: the username of the IG user to scrape
        """
        username = username.lstrip('@')
        endpoint = f"https://i.instagram.com/api/v1/users/web_profile_info/?username={username}"
        response = self.request(endpoint)
        return self._wrap(username, response, InstaUser)

    def _wrap(self, page: str, response: Response, Wrapper: Type[InstaPage]) -> InstaPage:
        """Validates and wraps the API response from an Instagram page
        """
        page = f'user @{page}' if Wrapper is InstaUser else f'hashtag #{page}'
        if response.ok:
            try:
                return Wrapper(response.json(), self)
            except JSONDecodeError as e:
                raise RuntimeError(f'Unable to scrape Instagram {page}') from e
        else:
            try:
                error = response.json()
            except JSONDecodeError:
                error = response.reason
            raise RuntimeError(
                'Failed to scrape Instagram {page}\nResponse: [{code}] -- {e}'.format(
                    page=page, code=response.status_code, e=error
                )
            )

    def get_username(self, user_id: Union[int, str]) -> str:
        """Retrieves the Instagram username for the user with the provided ``user_id``

        .. tip:: Use this with :meth:`get_user` to scrape by ``user_id``::

           >> user_id = 51276430399
           >> username = insta.get_username(user_id)
           >> user = insta.get_user(username)
           >> print(user.posts[0])

           Post 2981866202934977614 by @dailykittenig on 2022-11-29 01:44:37

        :param user_id: the id of the Instagram user to retrieve the username of
        """
        endpoint = f"https://i.instagram.com/api/v1/users/{user_id}/info"
        response = self.request(endpoint)
        if response.ok:
            return response.json().get('user', {}).get('username', '')
        else:
            raise RuntimeError(f"Failed to retrieve info for Instagram user with id {user_id}")

    def download_post(self, post: InstaPost, filepath: Optional[str] = None) -> bool:
        """Downloads the media from an Instagram post

        :param post: the :class:`~.InstaPost` of the post to download
        :param filepath: the path to save the downloaded media; if ``None``, saves to the :attr:`~DOWNLOAD_DIR`
        """
        if post.is_carousel:
            for child in post.children:
                self.download_post(child)
            return post.is_downloaded

        response = self.request(post.media_url)
        if not response.ok:
            print(f'Failed to download post {post.permalink} by {post.owner["username"]}')
            return False

        filepath = filepath if filepath else os.path.join(self.DOWNLOAD_DIR, post.filename)
        with open(filepath, 'wb') as f:
            f.write(response.content)

        print(f'Downloaded post {post.permalink} by {post.owner["username"]} to {filepath}')
        post.filepath = filepath
        return True

    @property
    def headers(self) -> Dict:
        """Headers to use in :meth:`~.request`"""
        return {'User-Agent': self.user_agent, }

    @property
    def cookies(self) -> Dict:
        """Cookies to use in :meth:`~.request`"""
        return {'sessionid': self.session_id, }



from __future__ import annotations
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Dict, Optional, TYPE_CHECKING, List
from . import InstaPost

if TYPE_CHECKING:
    from . import InstaClient


class InstaPage(ABC):

    """Abstract wrapper class for wrapping API responses from Instagram pages"""

    def __init__(self, data: Dict, client: Optional[InstaClient] = None):
        """Initialize an :class:`InstaPage`

        Used to wrap responses from endpoints that contain Instagram post data,
        like Instagram user profiles and Instagram hashtag searches

        :param data: the API response JSON to use as source data
        :param client: the :class:`~.InstaClient` to use; required for :meth:`~.get_more_posts`
        """
        self.data = data
        self.client = client
        self._posts = []

    @abstractmethod
    def __str__(self) -> str:
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the Instagram page"""
        pass

    @property
    @abstractmethod
    def page_data(self) -> Dict:
        """Data about the Instagram page itself"""
        pass

    @property
    @abstractmethod
    def media_data(self) -> Dict:
        """Data about posts on the Instagram page"""
        pass

    @property
    def id(self) -> int:
        """ID of the Instagram page"""
        return int(self.page_data.get('id', -1))

    @property
    def posts(self) -> List[InstaPost]:
        """Posts that have been scraped from the Instagram page

        To retrieve the next page of posts, call :meth:`get_more_posts`

        :returns: the page's posts as :class:`~.InstaPost` objects
        """
        if not self._posts:
            if edges := self.media_data.get('edges'):
                self._posts = [InstaPost(edge['node'], self.client) for edge in edges]
        return self._posts

    def get_more_posts(self) -> bool:
        """Requests the next page of posts from the :class:`InstaPage`

        If the page :attr:`~.has_more_posts`, they'll be added to the :attr:`~.posts` list

        :returns: ``True`` if the request was successful, otherwise ``False``
        """
        if not self.client:
            raise AttributeError("Must provide an InstaClient to scrape with")

        if not self.has_more_posts:
            print("All posts have already been scraped")
            return False

        if not (next_page := self._get_next_page()):
            print("Unable to retrieve the next page of posts")
            return False

        self.media_page_info.update(next_page.media_page_info)
        self._posts.extend(next_page.posts)
        return True

    @abstractmethod
    def _get_next_page(self) -> Optional[InstaPage]:
        """Makes the request for the next page of posts; wraps the response if successful"""
        pass

    @property
    def has_more_posts(self) -> bool:
        """Returns ``True`` if more posts can be scraped using :meth:`~.get_more_posts`"""
        return self.media_page_info.get('has_next_page')

    @property
    def end_cursor(self) -> str:
        """Cursor used in request by :meth:`~.get_more_posts`"""
        return self.media_page_info.get('end_cursor', '').strip('=')

    @property
    def media_page_info(self) -> Dict:
        return self.media_data.get('page_info', {})


class InstaUser(InstaPage):

    """API response wrapper for an Instagram user's profile"""

    def __init__(self, data: Dict, client: Optional[InstaClient] = None):
        """Initialize an :class:`InstaUser`

        :param data: the API response from :meth:`~.get_user`
        :param client: the :class:`~.InstaClient` to use
        """
        super().__init__(data, client)

    def __str__(self) -> str:
        return f"Instagram User: @{self.name}"

    @property
    def name(self) -> str:
        return self.page_data.get('username')

    @property
    def page_data(self) -> Dict:
        return self.data.get('data', {}).get('user', {})

    @property
    def media_data(self) -> Dict:
        return self.page_data.get('edge_owner_to_timeline_media', {'edges': []})

    def _get_next_page(self) -> Optional[InstaPage]:
        endpoint = 'https://www.instagram.com/graphql/query/?query_hash=8c2a529969ee035a5063f2fc8602a0fd' + \
                   f'&variables=%7B%22id%22%3A%22{self.id}%22%2C%22first%22%3A12%2C%22' + \
                   f'after%22%3A%22{self.end_cursor}%3D%3D%22%7D'
        response = self.client.request(endpoint)
        if not response.ok:
            return None
        try:
            return InstaUser(response.json())
        except Exception as e:
            raise RuntimeError('Failed to get more posts') from e


class Hashtag(InstaPage):

    """API response wrapper for an Instagram hashtag"""

    def __init__(self, data: Dict, client: Optional[InstaClient] = None):
        """Initialize a :class:`Hashtag`

        :param data: the API response from :meth:`~.get_hashtag`
        :param client: the :class:`~.InstaClient` to use
        """
        if (data := data.get('graphql', {}).get('hashtag')) is None:
            raise ValueError(f"Hashtag response data is missing")

        super().__init__(data, client)
        self._top_posts = []

    def __str__(self) -> str:
        return f"Instagram Hashtag: {self.name}"

    @property
    def name(self) -> str:
        return "#" + self.page_data.get('name')

    @property
    def page_data(self) -> Dict:
        return self.data

    @property
    def media_data(self) -> Dict:
        return self.page_data.get('edge_hashtag_to_media', {'count': 0, 'edges': []})

    @cached_property
    def top_posts(self) -> List[InstaPost]:
        return [InstaPost(edge['node'], self.client) for edge in self.top_media_data['edges']]

    @property
    def top_media_data(self) -> Dict:
        return self.page_data.get("edge_hashtag_to_top_posts", {"edges": []})

    def _get_next_page(self) -> Optional[InstaPage]:
        return self.client.get_hashtag(self.name, max_id=self.end_cursor)



import os
import requests
from pathlib import Path
from typing import Optional


AGENTS = ['Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.5005.63 Safari/537.36']


def get_agents() -> list:
    """Scrapes a list of user agents. Returns a default list if the scrape fails.

    ..  note:: Deprecated since 2.0.0b13, but might be useful when new endpoint gets patched
    """
    if (response := requests.get('https://www.whatismybrowser.com/guides/the-latest-user-agent/chrome')).ok:
        section = response.text.split('<h2>Latest Chrome on Windows 10 User Agents</h2>')[1]
        raw_agents = section.split('code\">')[1:]
        agents = [agent.split('<')[0] for agent in raw_agents]
        for a in agents:
            if a not in AGENTS:
                AGENTS.append(a)
    # If function fails, will still return the hardcoded list
    return AGENTS


def get_agent(index: int = 0) -> str:
    """Returns a single user agent string from the specified index of the AGENTS list

    ..  note:: Deprecated since 2.0.0b13, but might be useful when new endpoint gets patched
    """
    return get_agents()[index]  # Specify index only if you hardcode more than 1


def get_proxies(env_key) -> Optional[dict]:
    """Retrieve proxies from an environment variable"""
    if env_key:
        return {
            "http": os.environ[env_key],
            "https": os.environ[env_key]
        }
    return None


def get_root() -> Path:
    return Path(__file__).parent


def get_filepath(filename: str, filetype: str = 'txt') -> str:
    return os.path.join(get_root(), filename) + '.' + filetype



from typing import Optional, List, Dict
from . import utils, TweetClient, InstaClient, InstaPost, Profile


class InstaTweet:

    """Uses the settings from a Profile to do the actual InstaTweeting

    You might be wondering, what's InstaTweeting? According to TDK Dictionary:

        .. admonition:: **InstaTweet** (`verb`):
           :class: instatweet

           To load a :class:`~.Profile` 🠖 scrape :attr:`~.posts` from its Instagram pages
           🠖 :meth:`~.download_post` & :meth:`~.send_tweet` for any new content
           🠖 update the :attr:`~.page_map`
           🠖 :meth:`~.save` the profile if it :attr:`~.exists`

            .. admonition:: **Example Sentence**
               :class: example

               Oh, you lost 700 Twitter followers after you shared your IG post? Well maybe if people actually saw the
               picture and not just the caption your tweet would've been less creepy. You should've InstaTweeted it.

    """

    def __init__(self, profile: Profile):
        """Initializes :class:`~.InstaTweet` using a fully configured :class:`~.Profile`

        The :class:`~.Profile` will be used to initialize an :class:`~.InstaClient` and :class:`~.TweetClient`

        .. note:: InstaTweet won't :meth:`~.validate` the Profile settings until you call :meth:`~.start`

        :param profile: the :class:`~.Profile` to use for InstaTweeting
        """
        self.profile = profile
        self.proxies = self.get_proxies()
        self.insta = self.get_insta_client()
        self.twitter = self.get_tweet_client()

    @classmethod
    def load(cls, profile_name: str, local: bool = True) -> "InstaTweet":
        """Loads a profile by name

        :param profile_name: name of the Profile to load
        :param local: whether the profile is saved locally (default) or remotely on a database
        """
        return cls(profile=Profile.load(name=profile_name, local=local))

    def get_proxies(self) -> Optional[Dict]:
        """Retrieve proxies using the loaded Profile's :attr:`~Profile.proxy_key`"""
        return utils.get_proxies(
            env_key=self.profile.proxy_key
        )

    def get_insta_client(self) -> InstaClient:
        """Initializes an :class:`~.InstaClient` using the loaded :class:`~.Profile` settings"""
        return InstaClient(
            session_id=self.profile.session_id,
            user_agent=self.profile.user_agent,
            proxies=self.proxies
        )

    def get_tweet_client(self) -> TweetClient:
        """Initializes an :class:`~.TweetClient` using the loaded :class:`~.Profile` settings"""
        return TweetClient(
            profile=self.profile,
            proxies=self.proxies
        )

    def start(self, max_posts: int = 12) -> None:
        """InstaTweets all pages that have been added to the loaded :class:`~.Profile`

        The most recent posts from each page will be scraped, then compared to the ``scraped``
        list in the :attr:`~.PAGE_MAPPING` to determine which are new.

        Up to ``max_posts`` new posts from each page will then be downloaded and tweeted

        .. note:: If ``InstaTweet`` fails to :meth:`~.download_post` or :meth:`~.send_tweet`,
           the :attr:`~.PAGE_MAPPING` won't be updated

           * This ensures that failed repost attempts are retried in the next call to :meth:`~start`

           If a save file for the Profile already :attr:`~.exists`, successful reposts
           will trigger a call to :meth:`~.save`

        :param max_posts: the maximum number of new posts to download and tweet per page
        """
        profile = self.profile
        profile.validate()

        print(f'Starting InstaTweet for Profile: {profile.name}')

        for page in profile.page_map:
            page_name = page if page.startswith("#") else "@" + page

            if not (new_posts := self.get_new_posts(page)):
                print(f'No posts to tweet for {page_name}')
                continue

            print(f'There are {len(new_posts)} posts to tweet for {page_name}')
            hashtags = profile.get_hashtags_for(page)

            for post in new_posts[:max_posts]:
                self.insta.download_post(post)
                if not post.is_downloaded:
                    continue

                tweeted = self.twitter.send_tweet(post, hashtags)
                if not tweeted:
                    continue

                profile.get_scraped_from(page).append(post.id)
                profile.get_tweets_for(page).append(post.tweet_data)

                if profile.exists:
                    profile.save(alert=False)

            print(f'Finished insta-tweeting for {page_name}')

        print(f'All pages have been insta-tweeted')

    def get_new_posts(self, insta_page: str) -> Optional[List[InstaPost]]:
        """Scrapes recent posts from an Instagram page and returns all posts that haven't been tweeted yet

        **NOTE:**  If a page's ``scraped`` list is empty, no posts will be returned.

        Instead, the page is "initialized" as follows:

        * The ``scraped`` list will be populated with the ID's from the most recent posts
        * These IDs are then used in future method calls to determine which posts to tweet

        :param insta_page: the Instagram page to scrape posts from
        :return: a list of posts that haven't been tweeted yet, or nothing at all (if page is only initialized)

        """
        print(f'Checking posts from {insta_page}')
        scraped_posts = self.profile.get_scraped_from(insta_page)
        page = self.insta.scrape(insta_page)

        if scraped_posts:
            new_posts = [post for post in page.posts if post.id not in scraped_posts]
            return sorted(new_posts, key=lambda post: post.timestamp)
        else:
            scraped_posts.extend(post.id for post in page.posts)
            print(f'Initialized {page}')
            return None




from __future__ import annotations
import os
import pickle
import InstaTweet

from sqlalchemy import create_engine, Column, String, LargeBinary
from sqlalchemy.orm import sessionmaker, scoped_session, Query
from sqlalchemy.ext.declarative import declarative_base


DATABASE_URL = os.getenv('DATABASE_URL', '').replace('postgres://', 'postgresql://', 1)
"""The Database URL to use, obtained from the ``DATABASE_URL`` environment variable"""

Base = declarative_base()
"""Base for creating tables"""


class Profiles(Base):
    """Database table used for storing :class:`~.Profile` settings

    The table currently has only 2 fields, for the :attr:`~.Profile.name` and pickle bytes of the profile
    """
    __tablename__ = 'profiles'
    name = Column(String, primary_key=True)  #: The :class:`~.Profile` name
    config = Column(LargeBinary)  #: The pickle bytes from :meth:`.Profile.to_pickle()`

    def __repr__(self):
        return "<Profiles(name='{}')>".format(self.name)


class DBConnection:

    """Database Connection class with context management ooh wow

    Uses ``SQLAlchemy`` to connect and interact with the database specified by :attr:`DATABASE_URL` environment variable

    **Sample Usage**::

        def poop_check():
            with DBConnection() as db:
                if db.query_profile(name="POOP").first():
                    raise FileExistsError('DELETE THIS NEPHEW......')
    """

    SESSION = None
    """The currently active session; closed on object exit
    
    :type: :class:`~.sqlalchemy.orm.scoping.scoped_session`
    """

    ENGINE = None
    """The engine for the currently set :attr:`DATABASE_URL`; reused after first connection
    
    :type: :class:`~.sqlalchemy.engine.base.Engine`
    """

    def __enter__(self):
        if not DATABASE_URL:
            raise EnvironmentError('Must set the DATABASE_URL environment variable')

        if not self.ENGINE:
            engine = create_engine(DATABASE_URL, echo=False)
            Base.metadata.create_all(engine)
            DBConnection.ENGINE = engine

        if not self.SESSION:
            self.connect()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        DBConnection.SESSION = None

    @staticmethod
    def connect() -> None:
        """Creates a :class:`~.sqlalchemy.orm.scoping.scoped_session` and assigns it to :attr:`DBConnection.SESSION`"""
        DBConnection.SESSION = scoped_session(sessionmaker(bind=DBConnection.ENGINE))

    def query_profile(self, name: str) -> Query:
        """Queries the database for a :class:`~.Profile` by its name

        :param name: the profile name (ie. the :attr:`.Profile.name`)
        :returns: the :class:`~sqlalchemy.orm.Query` NOT the :class:`~.Profile`
        """
        return self.SESSION.query(Profiles).filter_by(name=name)

    def load_profile(self, name: str) -> InstaTweet.Profile:
        """Loads a profile from the database by name

        :param name: the profile name (ie. the :attr:`.Profile.name`)
        :raises LookupError: if the database has no profile saved with the specified name
        """
        if profile := self.query_profile(name).first():
            return pickle.loads(profile.config)
        else:
            raise LookupError(f"No database profile found with the name {name}")

    def save_profile(self, profile: InstaTweet.Profile, alert: bool = True) -> bool:
        """Saves a :class:`~.Profile` to the database by either updating an existing row or inserting a new one

        :param profile: the :class:`~.Profile` to save
        :param alert: if ``True``, will print a message upon successfully saving
        """
        if (db_profile := self.query_profile(profile.name)).first():
            if profile.local:  # Would only happen if saved directly through this class [not recommended]
                raise ResourceWarning(  # Save through the Profile itself to avoid this
                    f"Database profile with the name {profile.name} already exists")
            db_profile.update({'config': profile.to_pickle()})
        else:
            new_profile = Profiles(name=profile.name, config=profile.to_pickle())
            self.SESSION.add(new_profile)

        self.SESSION.commit()

        if alert:
            print(f"Saved Database Profile: {profile.name}")

        return True

    def delete_profile(self, name: str, alert: bool = True) -> bool:
        """Deletes a :class:`~.Profile` from the database by name

        :param name: the profile name (ie. the :attr:`.Profile.name`)
        :param alert: if ``True``, will print a message upon successfully deleting
        """
        if not (profile := self.query_profile(name).first()):
            raise LookupError(f"No database profile found with the name {name}")

        profile.delete()
        self.SESSION.commit()

        if alert:
            print(f'Deleted Database Profile: {name}')
        return True



import os
from pathlib import Path
from setuptools import setup

LONG_DESCRIPTION_SRC = 'README_PyPi.rst'


def read(file):
    with open(os.path.abspath(file), 'r', encoding='utf-8') as f:
        return f.read()


# Parse version
init = Path(__file__).parent.joinpath("InstaTweet", "__init__.py")
for line in init.read_text().split("\n"):
    if line.startswith("__version__ ="):
        break
version = line.split(" = ")[-1].strip('"')


setup(
    name='insta-tweet',
    packages=['InstaTweet'],
    version=version,
    license='MIT',
    description='Automatically Repost Content From Instagram to Twitter',
    long_description=read(LONG_DESCRIPTION_SRC),
    long_description_content_type="text/x-rst; charset=UTF-8",
    author='Adam Korn',
    author_email='hello@dailykitten.net',
    url='https://www.github.com/TDKorn/insta-tweet/',
    download_url=f"https://github.com/TDKorn/insta-tweet/tarball/master/",
    keywords=['instagram', 'twitter', 'api', 'instagram api', 'twitter api', 'repost', 'instagram repost', 'reposter'],
    install_requires=[
        "requests",
        "tweepy",
        "sqlalchemy",
    ],
)



from __future__ import annotations

import os
import random
import InstaTweet

from tweepy import OAuth1UserHandler, API, Media, TweepyException
from typing import Union, Optional
from . import InstaPost


class TweetClient:

    MAX_HASHTAGS = 5
    DEFAULT_KEYS = {
        'Consumer Key': 'string',
        'Consumer Secret': 'string',
        'Access Token': 'string',
        'Token Secret': 'string'
    }

    def __init__(self, profile: InstaTweet.Profile, proxies: dict = None):
        """Initialize TweetClient using a :class:`~.Profile`

        Basically just a wrapper for tweepy.
        It uses the settings of a profile to initialize the API and send tweets

        :param profile: the profile to use when initializing a :class:`tweepy.API` object
        :param proxies: optional proxies to use when making API requests
        """
        self.profile = profile
        self.proxies = proxies
        self.api = self.get_api()

    def get_api(self) -> API:
        """Initializes a :class:`~.tweepy.API` object using the API keys of the loaded :class:`~.Profile`"""
        return API(
            auth=self.get_oauth(self.profile.twitter_keys),
            user_agent=self.profile.user_agent,
            proxy=self.proxies
        )

    @staticmethod
    def get_oauth(api_keys: dict) -> OAuth1UserHandler:
        """Initializes and returns an :class:`~.OAuth1UserHandler` object from tweepy using the specified API keys

        :param api_keys: Twitter developer API keys with v1.1 endpoint access
        """
        if missing_keys := [key for key in TweetClient.DEFAULT_KEYS if key not in api_keys]:
            raise KeyError(
                f"Missing the following Twitter Keys: {missing_keys}"
            )
        if bad_keys := [key for key in TweetClient.DEFAULT_KEYS if not api_keys[key] or api_keys[key] == 'string']:
            raise ValueError(
                f"Invalid values for the following Twitter keys: {bad_keys}"
            )
        return OAuth1UserHandler(
            consumer_key=api_keys['Consumer Key'],
            consumer_secret=api_keys['Consumer Secret'],
            access_token=api_keys['Access Token'],
            access_token_secret=api_keys['Token Secret']
        )

    def send_tweet(self, post: InstaPost, hashtags: Optional[list[str]] = None) -> bool:
        """Composes and sends a Tweet using an already-downloaded Instagram post

        .. admonition:: How Tweets are Sent
           :class: instatweet

           The :attr:`.InstaPost.filepath` -- set by :meth:`~.download_post` -- is used as the media source

           The body of the tweet is then generated by :meth:`~build_tweet`

        :param post: the post to tweet
        :param hashtags: a list of hashtags to randomly chose from and include in the tweet
        """
        if not post.is_downloaded:
            raise FileNotFoundError('Post must be downloaded first')

        media_ids = self.upload_media(post)
        if not isinstance(media_ids, list):
            return False

        try:
            tweet = self.api.update_status(
                status=self.build_tweet(post, hashtags),
                media_ids=media_ids,
            )
        except TweepyException as e:
            print('Failed to send tweet for {}:\nResponse: {}'.format(post, e))
            return False

        print(f'Sent tweet for {post}')
        return post.add_tweet_data(tweet)

    def upload_media(self, post: InstaPost) -> Union[list, bool]:
        """Uploads the media from an already-downloaded Instagram post to Twitter

        .. note:: If the post is a carousel, only the first 4 photos/videos will be uploaded

        :param post: the Instagram post to use as the media source
        :return: the list of uploaded media ids (if API upload was successful) or ``False``
        """
        if not post.is_downloaded:
            raise FileNotFoundError('Post must be downloaded first')

        content = post.children[:4] if post.is_carousel else [post]
        media_ids = []

        for media in content:
            media_upload = self.api.media_upload(
                filename=media.filepath,
                media_category='TWEET_VIDEO' if media.is_video else 'TWEET_IMAGE',
                wait_for_async_finalize=True,
                chunked=True)

            if hasattr(media_upload, 'processing_info'):
                if media_upload.processing_info['state'] != 'succeeded':
                    print(f'Failed to upload media to Twitter for {media}')
                    return False

            media_ids.append(str(media_upload.media_id))

        print(f'Successfully uploaded media to Twitter for {post}')
        return media_ids

    def build_tweet(self, post: InstaPost, hashtags: Optional[list[str]] = None) -> str:
        """Uses an :class:`~.InstaPost` to build the body text of a tweet

        .. admonition:: How Tweets are Composed
           :class: instatweet

           - The :attr:`~.InstaPost.caption` is used as a starting point
           - If you :meth:`~.add_hashtags` for the page, it will randomly :meth:`~pick_hashtags` to include
           - Lastly, the post's :attr:`~.InstaPost.permalink` is added to the end

        **Example**::

         >> post = instatweet.insta.get_user("dailykittenig").posts[0]
         >> tweet = instatweet.twitter.build_tweet(post)
         >> print(tweet)

         carousel support yuh
         #kitten #kittycat #catlover #petstagram #animals

         https://www.instagram.com/p/Cjl3UWBOd8k

        :param post: the post being tweeted
        :param hashtags: optional list of hashtags to randomly pick from and include
        :return: the text to use for the tweet
        """
        tags = self.pick_hashtags(hashtags)
        caption = post.caption.strip().replace('@', '@/')  # Avoid tagging randos on Twitter
        characters = 280 - len(tags) - len(post.permalink) - 2
        tweet = "{text}\n{hashtags}\n{link}".format(
            text=caption[:characters],
            hashtags=tags,
            link=post.permalink
        )
        return tweet

    @staticmethod
    def pick_hashtags(hashtags: list[str]) -> str:
        """Randomly picks hashtags from the provided list and returns them as a single string

        The number of hashtags chosen will either be 1 less than the length of the list (to avoid using the same tags
        in every tweet), or the value of :attr:`~.MAX_HASHTAGS`, whichever is smaller

        :param hashtags: a list of hashtags to randomly choose from

        :Example:

            ::

               from InstaTweet import TweetClient

               >> TweetClient.pick_hashtags(['cat','dog','woof'])
               "#woof #cat\\n"

        .. note:: A newline is added to help with formatting & character counting in :meth:`~.build_tweet`

        """
        if not hashtags:
            return ''
        if not isinstance(hashtags, list):
            raise TypeError('Provide a list of hashtags')

        num_hashtags = min(len(hashtags) - 1, TweetClient.MAX_HASHTAGS)  # Pick at most MAX_HASHTAGS
        random_hashtags = random.sample(hashtags, max(1, num_hashtags))  # Pick at least 1

        return ' '.join(f'#{hashtag}' for hashtag in random_hashtags) + '\n'
